import json
import os
from typing import Dict, List, TextIO, Tuple

import yaml


def _get_format(template_file: str) -> str:
    _, extension = os.path.splitext(template_file)
    return extension.lower()[1:]


def _parse_template(content: TextIO, format_name: str) -> Dict:
    yaml.add_multi_constructor("", any_constructor, Loader=yaml.SafeLoader)
    if format_name in ["yml", "yaml"]:
        return yaml.safe_load(content)
    if format_name == "json":
        return json.load(content)


def any_constructor(loader, _, node):
    if isinstance(node, yaml.MappingNode):
        return loader.construct_mapping(node)
    if isinstance(node, yaml.SequenceNode):
        return loader.construct_sequence(node)
    return loader.construct_scalar(node)


def get_endpoints(
    template_file: str, exclude: List[str], proxy_lambdas: List[str]
) -> Tuple[List[Dict], List[str]]:
    """parse Cloud Fomation Template to extract info about lambdas"""
    _resources = {}
    layers = []

    with open(template_file, "r") as tf:
        format_name = _get_format(template_file)
        template = _parse_template(tf, format_name)
        resources = template["Resources"]
        for resource, resource_data in resources.items():
            if "AWS::Serverless::Function" in resource_data["Type"]:
                _resources[resource] = resource_data
            elif "AWS::Serverless::LayerVersion" in resource_data["Type"]:
                layers.append(resource_data["Properties"]["ContentUri"])

    lambdas, proxy_lambdas = process_lambdas_resources(
        _resources, exclude, proxy_lambdas
    )

    return lambdas, layers, proxy_lambdas


def process_lambdas_resources(
    resources: Dict, exclude: List[str], proxy_lambdas: List[str]
) -> List[Dict]:
    lambdas = []
    p_lambdas = []
    for lambda_name, data in resources.items():
        if lambda_name not in exclude:
            code_uri = data["Properties"]["CodeUri"].replace("/", ".")
            if not code_uri.endswith("."):
                code_uri += "."
            handler = f"{code_uri}{data['Properties']['Handler']}"
            events = data["Properties"].get("Events")
            if lambda_name in proxy_lambdas:
                p_lambdas.append({"name": lambda_name, "handler": handler})
                continue
            if not events:
                # mean it is authorizer or smth relative
                continue
            lambdas.extend(get_lambdas_endpoints(events, lambda_name, handler))
    return lambdas, p_lambdas


def get_lambdas_endpoints(
    events: Dict, lambda_name: str, handler_str: str
) -> List[Dict]:
    endpoints = []
    for _, event in events.items():
        _type = event["Type"]
        if _type != "Api":
            continue
        method = event["Properties"]["Method"].lower()
        if method == "options":
            # this is for CORS, we not insteresting
            continue
        endpoint_info = {
            "name": lambda_name,
            "method": method,
            "endpoint": event["Properties"]["Path"],
            "handler": handler_str,
        }
        endpoints.append(endpoint_info)
    return endpoints



import os
import sys
from typing import Dict, List, Optional

import uvicorn
from fastapi import FastAPI

from lof.code_gen import create_temp_app_folder, generate_app
from lof.env_vars import path_for_env_file, set_env_variables
from lof.errors import NoTemplate, NoVariablesFile
from lof.generator import create_middleware, create_route
from lof.parser import get_endpoints
from lof.providers.aws import Context


def validate_paths(template_file: str, variables_file: str) -> None:
    if "/" not in template_file:
        template_file = os.path.join(os.getcwd(), template_file)
    if not os.path.isfile(template_file):
        raise NoTemplate(template_file)

    if variables_file and not os.path.isfile(variables_file):
        raise NoVariablesFile(variables_file)


def add_template_path_to_python_paths(template_file: str, layers: List[str]) -> None:
    lambdas_base_dir = os.path.dirname(template_file)
    sys.path.insert(0, lambdas_base_dir)
    for layer in layers:
        sys.path.insert(0, layer)


def run_fast_api_server(lambdas: List[Dict], proxy_lambdas: List[Dict]) -> FastAPI:

    app = FastAPI()
    app.context = Context

    create_middleware(proxy_lambdas=proxy_lambdas, app=app)

    for _lambda in lambdas:
        create_route(
            endpoint=_lambda["endpoint"],
            method=_lambda["method"],
            handler=_lambda["handler"],
            app=app,
            lambda_name=_lambda["name"],
        )
    return app


def run_multiple_workers_app(
    lambdas: List[Dict],
    port: int,
    host: str,
    workers: int,
    debug: bool,
    reload: bool,
    variables_file: str,
    proxy_lambdas: List[Dict],
) -> None:
    lof_dir = create_temp_app_folder()
    env_file = None
    if variables_file:
        env_file = path_for_env_file(variables_file, lof_dir)
    generate_app(lambdas=lambdas, lof_dir=lof_dir, proxy_lambdas=proxy_lambdas)
    uvicorn.run(
        "lof_app.main:app",
        host=host,
        port=port,
        debug=debug,
        reload=reload,
        env_file=env_file,
        workers=workers,
    )


def runner(
    template_file: str,
    variables_file: Optional[str],
    exclude: Optional[List[str]],
    proxy_lambdas: Optional[List[str]],
    port: Optional[str] = 8000,
    host: Optional[str] = "0.0.0.0",
    workers: Optional[int] = 1,
    debug: Optional[bool] = True,
    reload: Optional[bool] = False,
) -> None:
    if workers <= 0:
        raise ValueError("Count of workers cannot be less when 1")

    validate_paths(template_file, variables_file)
    lambdas, layers, proxy_lambdas = get_endpoints(
        template_file, exclude, proxy_lambdas
    )

    add_template_path_to_python_paths(template_file, layers)

    if variables_file:
        set_env_variables(variables_file)

    if workers == 1 and not reload:

        app = run_fast_api_server(lambdas=lambdas, proxy_lambdas=proxy_lambdas)
        uvicorn.run(app, host=host, port=port, debug=debug)
    else:
        run_multiple_workers_app(
            lambdas=lambdas,
            port=port,
            host=host,
            workers=workers,
            debug=debug,
            reload=reload,
            variables_file=variables_file,
            proxy_lambdas=proxy_lambdas,
        )



import os
import sys
from typing import Dict, List

from jinja2 import Template

LOF_DIR = ".lof"

route_template = """
@app.{method}("{endpoint}")
async def {function_name}(request: Request, response: Response):
    my_module = importlib.import_module("{module}")
    _handler = getattr(my_module, "{handler}")
    app.event["pathParameters"] = request.path_params
    result = _handler(app.event, app.context(function_name="{lambda_name}"))
    status_code = result.get("statusCode") or result.get("status_code") or 200
    if result.get("body"):
        content = result.get("body")
    else:
        content = result
    for header, value in result.get("headers", {{}}).items():
        response.headers[header] = value
    if status_code == 204:
        response = Response(status_code=status_code)
    else:
        response = JSONResponse(
            content=content, status_code=status_code, headers=response.headers
        )
    return response
"""


def create_temp_app_folder() -> str:
    lof_path = os.path.join(os.getcwd(), LOF_DIR)
    os.makedirs(lof_path, exist_ok=True)
    lof_app = os.path.join(lof_path, "lof_app")
    os.makedirs(lof_app, exist_ok=True)
    sys.path.insert(0, lof_path)
    return lof_path


def get_function_name(method: str, endpoint: str) -> str:
    endpoint_name = endpoint.split("{")[0]
    for symbol in ["/", "-", "&", "$", "%", "."]:
        endpoint_name = endpoint_name.replace(symbol, "_")
    return f"{method}_{endpoint_name}"


def create_route(endpoint: str, method: str, handler: str, lambda_name: str) -> str:

    _handler = handler.split(".")
    module = ".".join(_handler[0:-1])
    _handler = _handler[-1]
    function_name = get_function_name(method, endpoint)

    return route_template.format(
        function_name=function_name,
        endpoint=endpoint,
        method=method,
        module=module,
        handler=_handler,
        lambda_name=lambda_name,
    )


def get_routes(lambdas: List[Dict]) -> List[str]:
    routes = []
    for _lambda in lambdas:
        route = create_route(
            endpoint=_lambda["endpoint"],
            method=_lambda["method"],
            handler=_lambda["handler"],
            lambda_name=_lambda["name"],
        )
        routes.append(route)
    return routes


def render_app_template(lambdas: List[Dict], proxy_lambdas: List[Dict]) -> str:
    current_dir = os.path.dirname(os.path.abspath(__file__))
    template_path = os.path.join(current_dir, "template.jinja2")
    with open(template_path) as file_:
        template = Template(file_.read())
    app_routes = get_routes(lambdas)
    app_code = template.render({"routes": app_routes, "proxy_lambdas": proxy_lambdas})
    return app_code


def save_app_code(app_code, lof_dir) -> None:
    with open(os.path.join(lof_dir, "lof_app", "main.py"), "w+") as main_module:
        main_module.write(app_code)


def generate_app(lambdas: List[Dict], lof_dir: str, proxy_lambdas: List[Dict]) -> str:
    app_code = render_app_template(lambdas, proxy_lambdas)
    save_app_code(app_code, lof_dir)



import importlib
from typing import Any, Callable, Dict, List

from fastapi import FastAPI, Request, Response
from fastapi.responses import JSONResponse

from lof.providers.aws import prepare_api_gateway_event


def create_route(
    endpoint: Dict, handler: Callable, lambda_name: str, method: str, app: FastAPI
):
    method = getattr(app, method)

    @method(endpoint)
    async def _(request: Request, response: Response):
        _handler = handler.split(".")
        module = ".".join(_handler[0:-1])
        _handler = _handler[-1]
        my_module = importlib.import_module(module)
        _handler = getattr(my_module, _handler)

        context = {"function_name": lambda_name}

        app.event["pathParameters"] = request.path_params

        response = prepare_response(
            _handler(app.event, app.context(**context)), response
        )
        return response


def prepare_response(lambda_handler_result: Any, response: Response) -> Response:
    result = lambda_handler_result
    status_code = result.get("statusCode") or result.get("status_code") or 200

    if result.get("body"):
        content = result.get("body")
    else:
        content = result
    for header, value in result.get("headers", {}).items():
        response.headers[header] = value
    if status_code == 204:
        response = Response(status_code=status_code)
    else:
        if isinstance(content, dict) or isinstance(content, list):
            response = JSONResponse(
                content=content, status_code=status_code, headers=response.headers
            )
        else:
            response = Response(
                content=content, status_code=status_code, headers=response.headers
            )
    return response


def create_middleware(proxy_lambdas: List[Dict], app: FastAPI):
    @app.middleware("http")
    async def _(request: Request, call_next):
        app.event = await prepare_api_gateway_event(request)

        response = None
        for _lambda in proxy_lambdas:
            handler = _lambda["handler"]
            _handler = handler.split(".")
            module = ".".join(_handler[0:-1])
            _handler = _handler[-1]
            my_module = importlib.import_module(module)
            _handler = getattr(my_module, _handler)
            context = {"function_name": _lambda["name"]}

            result = _handler(app.event, app.context(**context))

            if "auth" in _lambda["name"].lower():
                if "context" in result:
                    app.event["requestContext"]["authorizer"] = result["context"]
                else:
                    app.event["requestContext"]["authorizer"] = result
            else:
                app.event["requestContext"].update(result)
        else:
            response = await call_next(request)
        return response



import json
import os
from typing import Dict

from dotenv import load_dotenv


def set_env_variables(variables_file: str):
    if os.path.basename(variables_file).endswith(".json"):
        variables = get_variables(variables_file)
        if "Parameters" in variables:
            variables = variables["Parameters"]
        for variable, value in variables.items():
            os.environ[variable] = value
    elif os.path.basename(variables_file).startswith("."):
        load_dotenv(variables_file)


def get_variables(variables_file: str) -> Dict:
    if os.path.basename(variables_file).endswith(".json"):
        with open(variables_file, "r") as vf:
            variables = json.load(vf)
            return variables


def path_for_env_file(variables_file: str, lof_dir: str) -> str:
    file_path = variables_file
    if variables_file.endswith(".json"):
        file_path = create_temp_env_file(file_path, lof_dir)
    return file_path


def create_temp_env_file(variables_file: str, lof_dir: str) -> str:
    target_path = os.path.join(lof_dir, ".env")
    variables = get_variables(variables_file)
    variables_txt = "".join([f"{name}={value}\n" for name, value in variables.items()])
    with open(target_path, "w+") as env_file:
        env_file.write(variables_txt)
    return target_path



class LoFError(Exception):
    pass


class NoTemplate(LoFError):
    def __init__(self, path: str, *args, **kwargs):
        default_msg = f"No template file found on passed path {path}"
        if not args:
            args = (default_msg,)
        super().__init__(*args, **kwargs)


class NoVariablesFile(LoFError):
    def __init__(self, path: str, *args, **kwargs):
        default_msg = f"No variables file found on passed path {path}"
        if not args:
            args = (default_msg,)
        super().__init__(*args, **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lof', 'lof.providers']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.0.1,<4.0.0',
 'PyYAML>=5.4.1,<6.0.0',
 'fastapi>=0.66.0,<0.67.0',
 'python-dotenv>=0.18.0,<0.19.0',
 'typer>=0.3.2,<0.4.0',
 'uvicorn>=0.14.0,<0.15.0']

entry_points = \
{'console_scripts': ['lof = lof.cli:cli']}

setup_kwargs = {
    'name': 'lof',
    'version': '0.5.3',
    'description': 'AWS Lambdas on FastAPI it is a Cli utilit that up & run your Python lambdas in local env based on AWS CloudFormation Template (json or yaml)',
    'long_description': '\nAWS Lambdas on FastAPI - LoF\n----------------------------\n\n\n.. image:: https://img.shields.io/pypi/v/lof\n   :target: https://img.shields.io/pypi/v/lof\n   :alt: badge1\n \n.. image:: https://img.shields.io/pypi/l/lof\n   :target: https://img.shields.io/pypi/l/lof\n   :alt: badge2\n \n.. image:: https://img.shields.io/pypi/pyversions/lof\n   :target: https://img.shields.io/pypi/pyversions/lof\n   :alt: badge3\n\n.. image:: https://github.com/xnuinside/lof/actions/workflows/main.yml/badge.svg\n   :target: https://github.com/xnuinside/lof/actions/workflows/main.yml/badge.svg\n   :alt: workflow\n\n\nAWS Lambdas on FastAPI (LoF) is a command line tool that helps you fast & easy up & run your Python AWS Lambdas for tests and local development. \n\nPay attention, that this is only for **Python lambdas**.\n\nIt does not support any other programming languages.\n\nHow does it work?\n-----------------\n\nInstall\n^^^^^^^\n\n.. code-block:: bash\n\n\n       pip install lof\n\nNow run lof & provide to it path to your template yaml/json file.\nOr you can run it from source dir with template.yaml (/.json) without any args\n\nHow to use\n^^^^^^^^^^\n\n.. code-block:: bash\n\n\n       lof\n\n       # or if path custom\n\n       lof --template example/template.yaml\n\nYou can choose that lambdas exclude from run by passing their names:\n\n.. code-block:: bash\n\n\n       lof --template example/template.yaml --exclude=PostLambda2Function\n\nTo pass environment variables to Lambdas, use flag --env, you can pass variables in 2 formats - json format and \'.env\' format. Both files as examples presented in example/ folder\n\n.. code-block:: bash\n\n\n       lof --env=.env\n\n       # or \n\n       lof --env=vars.json\n\nAutorizer Lambda\n^^^^^^^^^^^^^^^^\n\nTo emulate behaviour of Authorizer lambda use flag --proxy-lambdas, where - LambdaAuthorizer must be changed to your lambda name from cloud formation template. Request will got through this proxy lambdas and only if everything ok it will call target lambda.\n\nSame as in API GAteway - all return values from proxy lambdas will update "requestContext" key in the event.\n\n.. code-block:: bash\n\n\n       lof --proxy-lambdas=LambdaAuthorizer\n\n       # or \n\n       lof --proxy-lambdas=LambdaAuthorizer,CORS # if you need to or more proxy lambdas\n\nOther settings\n--------------\n\n.. code-block:: text\n\n\n       Usage: lof [OPTIONS]\n\n       Options:\n     --template TEXT                 Path to AWS Code Deploy template with\n                                     lambdas  [default: template.yaml]\n\n     --env TEXT                      Path to file with environment variables\n     --exclude TEXT                  Exclude lambdas.FastAPI will not up & run\n                                     them. Pass as string with comma. Example:\n                                     PostTrafficHook,PretrafficHook.  [default: ]\n\n     --port INTEGER                  Port to run lof  [default: 8000]\n     --host TEXT                     Host to run lof  [default: 0.0.0.0]\n     --proxy-lambdas TEXT            Lambdas Names that must be used as Handlers\n                                     for request. For example, Authorizer Lambda\n                                     or CORS Lambds. Each time when you send\n                                     request to some lambda - it will go through\n                                     those lambdas and populate \'requestContext\'\n                                     in the event  [default: ]\n\n     --workers INTEGER               Count of unicorn workers to run.If you want\n                                     run more when 1 worker LoF will generate\n                                     temp FastAPI server code for your lambdas.\n                                     [default: 1]\n\n     --debug / --no-debug            Debug flag for Uvicorn  [default: True]\n     --reload / --no-reload          Reload flag for Uvicorn  [default: False]\n     --help                          Show this message and exit.\n\nThis mean, that lof will up & run all lambdas exclude this 2: PostTrafficHook & Roles\n\nDemo\n----\n\nwill be added soon\n\nExample\n-------\n\nTo try how LoF works you can use AWS CloudFormation template.yaml & Lambdas from example/ folder.\n\nIssues & features request\n-------------------------\n\nFill free to open Issues & report bugs. I will solve them as soon as possible. If you have any sugesstions or feature request - also fell free to open the issue.\n\nProblem Context\n---------------\n\nOn my current project I works a lot with AWS Lambdas & tries to up & run them with SAM local. \nAnd there is some issues especially when you work on the project with a big count of lambdas.\n\nSome of them:\n\n1) First of all it does not allow skip some lambdas form config\n2) It build lambdas inside each docker container so it takes significant time to build/rebuild & up all containers (and you need up all containers if you want to have fast integration tests)\n\nBoth points in the mix make impossible to use SAM in weak developers envs like VDI, for example.\n\nChangelog\n---------\n\n**v0.5.3**\n\n\n#. Fixed bunch of issues relative to missed pathParams in event & wrong authorizer context providing\n\n**v0.5.2**\n\n\n#. Bug fixies.\n\n**v0.5.0**\n\n\n#. Added AWS Context object. Part of values filed right now with mock values, values from config will be added in next versions.\n\n**v0.4.1**\nFeatures:\n\n\n#. Added option --proxy-lambdas where you can pass a list of lambdas that will be used as middleware request handlers. \n   For example, as Authorization Lambda or CORS lambda.\n\nBased on order in that you provided lambdas names in --proxy-lambdas option request will be send through them and populates same as on aws in "requestContext" field of the event.\nFor example, if you use --proxy-lambdas=CORS,Authorizer this mean request first of all will go to CORS lambda and if all ok (no raise errors) will got to Authorizer lambda and when to target lambda (endpoint that you call).\n\nFixes:\n\n\n#. Paths with symbols \'-.\' now does not cause issue during running with 1 and more worker.\n\n**v0.3.0**\n\n\n#. Added Possimility to run multiple workers with flag --workers. \n   This helpful if you need speed up your local server or some lambdas need to call another lambdas directly.\n#. Added flag --reload to cli if you want auto reload server when code changed (uvicor --reload)\n#. Added support for Cloud Formation templates in JSON\n\n**v0.2.3**\n\n\n#. Possibility to send port & host to start several instances in same time.\n\n**v0.2.2**\n\n\n#. README.md is updated\n#. Fixed Issue with lambdas in template, that does not have Events with Path (like S3 triggered lambdas)\n#. Fixed issue with status code 204 - now it returns correct answer with no failes.\n#. Added some tests\n\n**v0.2.1**\n\n\n#. Now LoF do not try/except lambdas errors\n\n**v0.2.0**\n\n\n#. Fixed status_code resend from lambda & JSON body response\n\n**v0.1.0**\n\n\n#. First version of Lambdas on FastApi. \n   Based on AWS CloudFormation template it\'s serve lambdas as FastAPI endpoints for local testing.\n',
    'author': 'Iuliia Volkova',
    'author_email': 'xnuinside@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/xnuinside/lof',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0',
}


setup(**setup_kwargs)



import json
from typing import Dict, Optional
from uuid import UUID, uuid4

from fastapi import Request
from pydantic import BaseModel, Field


class Context(BaseModel):

    function_name: str = "MockLambdaName"
    function_version: str = "latest"
    invoked_function_arn: str = "invoked_function_arn"
    memory_limit_in_mb: int = 500
    aws_request_id: UUID = Field(default=str(uuid4()))
    log_group_name: str = "log_group_name"
    log_stream_name: str = "log_stream_name"
    identity: UUID = Field(default=str(uuid4()))
    cognito_identity_id: UUID = Field(default=str(uuid4()))
    cognito_identity_pool_id: UUID = Field(default=str(uuid4()))

    @staticmethod
    def get_remaining_time_in_millis():
        return 4200


async def prepare_api_gateway_event(request: Request) -> Request:
    body = None
    try:
        body = await request.json()
    except Exception:
        pass
    multi_params = get_multi_value_params(request.url)
    headers = {}
    for header in request.headers.items():
        headers[header[0]] = header[1]
    event = {
        "resource": str(request.base_url),
        "path": str(request.url),
        "httpMethod": request.method,
        "requestContext": {
            "resourcePath": "/",
            "httpMethod": request.method,
            "path": str(request.base_url),
        },
        "headers": headers,
        "multiValueHeaders": None,
        "queryStringParameters": get_query_params(multi_params),
        "multiValueQueryStringParameters": multi_params,
        "pathParameters": request.path_params,
        "stageVariables": None,
        "body": json.dumps(body),
        "isBase64Encoded": False,
    }
    return event


def get_multi_value_params(url: str) -> Dict:
    """extract parmas from url for multiqueryParams"""
    url = str(url).split("/")[-1]
    base_part = str(url).split("/")[0]
    params = url.split("?")[-1]
    params = params.split("&")
    multi_query_params = {}
    if len(params) == 1 and params[0] == base_part:
        params = []
    for param in params:
        name, value = param.split("=")
        if not multi_query_params.get(name):
            multi_query_params[name] = [value]
        else:
            multi_query_params[name].append(value)
    if not multi_query_params:
        multi_query_params = None
    return multi_query_params


def get_query_params(multi_params: Optional[Dict]) -> Dict:
    params = {}
    if multi_params:
        for param in multi_params:
            params[param] = multi_params[param][-1]
        return params



from typing import Optional

import typer

from lof import runner


def main(
    template: str = typer.Option(
        "template.yaml", help="Path to AWS Code Deploy template with lambdas"
    ),
    env: str = typer.Option(None, help="Path to file with environment variables"),
    exclude: Optional[str] = typer.Option(
        "",
        help="Exclude lambdas."
        "FastAPI will not up & run them. Pass as string with comma. Example: PostTrafficHook,PretrafficHook.",
    ),
    port: Optional[int] = typer.Option(8000, help="Port to run lof"),
    host: Optional[str] = typer.Option("0.0.0.0", help="Host to run lof"),
    proxy_lambdas: Optional[str] = typer.Option(
        "",
        help="Lambdas Names that must be used as Handlers for request. "
        "For example, Authorizer Lambda or CORS Lambds. "
        "Each time when you send request to some lambda - it will go through those lambdas and populate "
        "'requestContext' in the event",
    ),
    workers: Optional[int] = typer.Option(
        1,
        help="Count of unicorn workers to run."
        "If you want run more when 1 worker LoF will generate temp FastAPI server code for your lambdas.",
    ),
    debug: Optional[bool] = typer.Option(True, help="Debug flag for Uvicorn"),
    reload: Optional[bool] = typer.Option(False, help="Reload flag for Uvicorn"),
):
    exclude = [i.strip() for i in exclude.split(",")]
    proxy_lambdas = [i.strip() for i in proxy_lambdas.split(",")]

    runner(
        template_file=template,
        variables_file=env,
        exclude=exclude,
        port=port,
        host=host,
        workers=workers,
        debug=debug,
        reload=reload,
        proxy_lambdas=proxy_lambdas,
    )


def cli():
    typer.run(main)
